from tsalib.ts import TS, DimVar, dummy_dvar, TupleSeq
from sympy import sympify, Symbol
import re

def _sexpr_to_ts (e, dummy_idx=-1, strict=False):
    '''
    A single string expression (sexpr) to Tensor Shape expressions (ts)
    Converts shorthand dummy/empty placeholders to dummy TSs
    '''
    if isinstance(e, TS):  
        t = e
    else: 
        assert isinstance(e, str)
        if e == '' or e =='_':  
            t = dummy_dvar(dummy_idx)
            dummy_idx += 1
        elif e == '^':
            #TODO: better way to handle '^' ?
            t = TS(Symbol(e))
        else: 
            #TODO: strict: check if all dim vars in e are previously declared?
            t = TS(sympify(e))

    return t, dummy_idx

def _sexprs_to_ts(exprs, strict=False):
    '''
    String expressions (sexprs) to Tensor Shape expressions (ts)
    Converts shorthand dummy/empty placeholders to dummy TSs
    Returns a tuple of TSs
    '''
    dummy_idx = 0
    res = []
    for e in exprs:
        t, dummy_idx = _sexpr_to_ts(e, dummy_idx, strict)
        res.append(t)

    #print (exprs, res)
    return tuple(res)


seq_re = r'\((.+)\)\*'

def _to_tuple (ss):
    '''
    :ss is shape string, e.g., 'btd' or 'b,t,d*2' or '(btd)*'
    :returns the shape representation in tuple/TupleSeq form
    '''
    if isinstance(ss, (list, tuple)):
        for s in ss: assert isinstance(s, (TS,int))
        return tuple(ss)
    elif isinstance(ss, TupleSeq):
        return ss
    elif isinstance(ss, str):
        #remove all whitespace characters
        ss = re.sub(r'\s+', '', ss)

        #check if shape corresponds to a sequence        
        is_seq = False
        m = re.search(seq_re, ss)
        if m is not None:  # ss = '(b,t,d)*'
            ss = m.groups()[0]  # ss = 'b,t,d'
            #print (f'groups: {m.groups()}') 
            is_seq = True

        if ',' in ss: exprs = ss.strip().split(',') #'b,t,d*2' -> ['b', 't', 'd*2']
        else: exprs = list(ss)              # 'btd' -> ['b','t','d']

        exprs = _sexprs_to_ts(exprs)
        for e in exprs:
            assert isinstance(e, TS)

        exprs = tuple(exprs)

        if is_seq:
            exprs = TupleSeq(exprs)

        return exprs

    else:
        raise ValueError('Unknown type of ss')

def check_int_tuple(s):
    #print(f'int tuple? {s}')
    for d in s:
        try: d = int(d)
        except:
            raise ValueError(f'Unable to resolve expression {d}')
def is_int_tuple(s):
    ret = all([isinstance(x, int) for x in s])
    return ret

def resolve_to_int_tuple(s):
    '''
    resolve non-int elements by casting to int or looking up their DimVar values
    '''
    res = []
    for d in s:
        try: 
            d = int(d)
            res.append(d)
        except:
            if isinstance(d, TS):
                e = d.exp
            elif isinstance(d, Symbol):
                e = d
            else:
                raise ValueError(f'Unknown item {d}: {type(d)}')
            
            r = DimVar.eval(e)
            #print('r is ', r)
            try: 
                r = int(r)
                res.append(r)
            except:
                raise ValueError(f'Unable to resolve {d}')

    return tuple(res)


def to_int_tuple(s):
    return tuple([int(d) for d in s])

def to_shape (src):
    '''
    src: 'b,t,h*d'
    Lookup each shorthand in cache. 
    returns: (B, T, H*D)
    '''
    return NotImplemented


from tsalib.ts import TS, dim_var, dummy_dvar, TupleSeq
from sympy import sympify, Integer
from .utils import _sexprs_to_ts, _to_tuple, check_int_tuple, resolve_to_int_tuple



def _view_transform (src, to, in_shape):
    '''
    View Transform
    src, to: Union[str, Tuple]
    :src is the current view of the tensor
    :to is the target view, may contain expressions over named dim variables
    :in_shape is the shape (list/tuple) of the source tensor 
    :returns the new size of the tensor after view transformation (backend independent)
    '''
    check_int_tuple(in_shape)

    src = _to_tuple(src)
    if (len(src) != len(in_shape)):
        print (f'{src}, {in_shape}')
        raise ValueError("Source TS does not match input tensor's shape")
    to = _to_tuple(to)
    #print (src, to)
    assert isinstance(in_shape, (list, tuple))
    assert isinstance(src, tuple)
    assert isinstance(to, tuple)

    #sub_map = [(d.e, Symbol(f'{i}')) for i, d in enumerate(src)]
    sub_map = [(d.exp, in_shape[i]) for i, d in enumerate(src)]
    out_shape = tuple([t.exp.subs(sub_map) if isinstance(t, TS) else int(t) for t in to])

    out_shape = resolve_to_int_tuple(out_shape)
    return out_shape

def view_transform (tfm, in_shape):
    '''
    View transform
    :tfm:str is the shorthand representation of the transform ('btd -> b,t*2,d//2')
    '''
    l, r = tfm.split('->')
    return _view_transform(l.strip(), r.strip(), in_shape)


def _permute_transform(src, to):
    '''
    Permute Transform
    src, to: Union[str, Tuple]

    :src is the current dimension arrangement, list of named dim variables
    :to is the target dimension arragement, list of named dim variables
    :returns the index tuple for the permutation (backend independent)
    '''
    src = _to_tuple(src)
    to = _to_tuple(to)
    assert isinstance(src, tuple)
    assert isinstance(to, tuple)

    assert len(src) == len(to), "Source and Target shapes for permutation are not same"

    sub_map = [(d.exp, i) for i, d in enumerate(src)]
    perm_indices = tuple([t.exp.subs(sub_map) for t in to])
    perm_indices = resolve_to_int_tuple(perm_indices)

    return perm_indices


def permute_transform (tfm):
    '''
    Permute transform
    :tfm: str
    :tfm is the shorthand representation of the transform (',t,d -> ,d,t')
    '''
    l, r = tfm.split('->')
    return _permute_transform(l.strip(), r.strip())

def _expansions_as_dict(expansions):
    if isinstance(expansions, list): #[(T, T*5), (D, D*4)]
        res = expansions
    else:
        assert isinstance(expansions, str) #'k->k*5,t->t*10'
        expansions = expansions.strip().split(',')
        res = []
        for ex in expansions:
            t = _sexprs_to_ts(ex.strip().split('->'))
            #print(t)
            res.append(t)
            #print (res)

    exp_map = {l: r for (l, r) in res}
    return exp_map


def expand_transform (src, expansions, in_shape):
    '''
    :src        (B, T, D) = (10, 20, 300)
    :expansions [(T, T*5), (D, D*4)]
    :returns the expansion shape tuple
    '''
    src = _to_tuple(src)
    exp_map = _expansions_as_dict(expansions)
    #exp_map = {_sexpr_to_ts(s)[0]: _sexpr_to_ts(t)[0] for (s, t) in (expansions)}
    sub_map = [(d.exp, in_shape[i]) for i, d in enumerate(src)] # (B, 10), (T, 20), (D, 300)

    #print (expansions, exp_map)

    res = []
    for k in src:
        if k not in exp_map: res.append(-1) #keep dim shape same
        else:
            v = exp_map[k]
            assert isinstance(v, TS)
            res.append(v.exp.subs(sub_map)) #(T*5) -> 100, (D*4) -> 1200

    res = tuple(res)
    res = resolve_to_int_tuple(res)
    return res

def reduce_dims (tfm):
    '''
    tfm: str, 'btd->b'
    '''
    src, to = tfm.split('->')
    src = _to_tuple(src.strip())
    to = _to_tuple(to.strip())

    assert isinstance(src, tuple)
    assert isinstance(to, tuple)

    drops = []
    #check src includes all dims in to
    for i, d in enumerate(src):
        if d not in to:
            drops.append(i)

    return tuple(drops)




def _join_transform (tlist, src, to):
    '''
    src: Union[str, TupleSeq]
    to: Union[str, tuple]
    src: (b,c,d)* , to: '^, b, c, d'
                  , to: 'b, 3*c, d'
    returns the dims shorthand for joining (backend independent)

    '''
    lhs = _to_tuple(src) #TupleSeq(B, C, D)
    rhs = _to_tuple(to) 

    assert isinstance(lhs, TupleSeq)
    assert isinstance(rhs, tuple)

    lhs = lhs.item() # (b, c, d)
    int1 = Integer(1) # substitute all dim shorthands by '1'

    sub_map = [(d.exp, int1) for d in lhs]
    dims = tuple([t.exp.subs(sub_map) for t in rhs]) # (^, 1, 1, 1)


    if len(rhs) == len(lhs): #concat, join by '*'
        #join_dims = (1,3*1,1)
        dims = ','.join(map(lambda x: '' if x == int1 else '*', dims))
        #dims = ',*,'
    elif len(rhs) == (len(lhs) + 1): #stack, join by '^'
        dims = ','.join(map(lambda x: '' if x == int1 else '^', dims))
        #dims = '^,,,'
    else:
        raise ValueError(f'Unable to join from {src} to {to}')

    return dims


def join_transform (tlist, tfm):
    '''
    Join transform
    :tlist: List[tensor]
    :tfm:str represents the transform "(b,c,d)* -> ^,b,c,d"
    returns the dims shorthand for joining (backend independent)

    '''
    l, r = tfm.split('->')
    return _join_transform(tlist, l.strip(), r.strip())



from .backend import get_backend_by_name, get_backend_for_tensor

def join (tlist, dims, backend=None):
    '''
    tlist: List[tensor], list of tensors
    dims: str = '..,^,...' or '..,*,...' 
    '''
    assert isinstance(tlist, list), "Can only group a list of tensors."
    assert len(tlist) > 1, "Can only group more than one tensors."

    if backend is not None:
        be = get_backend_by_name(backend)
    else:
        be = get_backend_for_tensor(tlist[0])

    if len(dims) > 1: assert ',' in dims, 'Separate dimensions by "," here.'

    out_shape = dims.strip().split(',')
    if '^' in out_shape:
        pos = out_shape.index('^')
        return be.stack(tlist, axis=pos)
    else:
        if '*' not in out_shape:
            assert pos != '-1', 'Group specification does not contain "^" or "*".'

        pos = out_shape.index('*')
        return be.concat(tlist, axis=pos)


def tfm_seq_decompose (tfms, tfm_names):
    '''
    Decompose a multi-step transform into basic (view, permute, expand) transforms
    tfms  'btd -> b,t,2,d//2 -> b,2,t,d//2'
    tfm_names 'vp' , i.e., view, then permute transform
    '''
    tfm_symbols = list(tfm_names) # ['v', 't']
    tfm_symbols_no_c = list(tfm_names.replace('c',''))
    tfm_list = [] # (trf symbol, trf_lhs, trf_rhs)

    if isinstance(tfms, str):

        shapes = [t.strip() for t in tfms.split('->')]
        assert len(shapes) >= 2, "Specify at least one transform, e.g., btd->dtb"
        assert len(tfm_symbols_no_c) == (len(shapes) - 1), "Num of transform descriptions and symbols do not match"

        shapes = [_to_tuple(s) for s in shapes]
        #for i, (l, r) in enumerate(zip(shapes[:-1], shapes[1:])):
        #    tfm_list.append((tfm_symbols[i], l, r) )

        curr_shape_pos = 0 #count current tfm's position (handle implicit contiguous)
        for sym in tfm_symbols:
            #contiguous transform
            if sym == 'c': 
                tfm_list.append((sym, None, None))
            else:
                l, r = shapes[curr_shape_pos: curr_shape_pos+2]
                tfm_list.append((sym, l, r))
                curr_shape_pos += 1

    elif isinstance(tfms, list):
        assert len(tfms) == len(tfm_symbols_no_c), "Num transformations {0} != transform symbols {1}".format(len(tfms),len(tfm_symbols_no_c))
        assert len(tfms) > 0, "No transformations given. Specify at least one transformation"
        curr_pos = 0 #count current tfm's position (handle implicit contiguous)
        for sym in tfm_symbols:
            if sym == 'c':   #contiguous transform
                tfm_list.append((sym, None, None))
            else:
                l, r = [t.strip() for t in tfms[curr_pos].split('->')]
                tfm_list.append((sym, l, r))
                curr_pos += 1

    else:
        assert False, "warp: wrong format for transforms. Specify either a string or a list."

    return tfm_list



def warp (x, tfms, tfm_names, backend=None, debug=False):
    '''
    Perform a multi-step transform on the tensor x
    tfms  'btd -> b,t,2,d//2 -> b,2,t,d//2 -> b,2,t,^n,d//2'
    tfm_names 'vp' [first (v)iew, then (p)ermute transform]
    backend    either a string('numpy', 'tf', 'torch') or the corresponding backend.<class>
    '''
    if backend is not None:
        be = get_backend_by_name(backend)
    else:
        be = get_backend_for_tensor(x)

    tfm_list = tfm_seq_decompose(tfms, tfm_names)

    #print (f'tfm list {tfm_list}')

    ret = x
    for sym, l, r in tfm_list:
        if debug:
            print(f'*** processing transform.. {sym}\n {l} -> {r}')
        if sym == 'v' or sym == 'r': #view transform
            new_shape = _view_transform(l, r, ret.shape)
            ret = be.view(x, new_shape)
        elif sym == 'p' or sym == 't':
            perm_indices = _permute_transform(l, r)
            ret = be.transpose(ret, perm_indices)
        elif sym == 'e':
            expand_shape = expand_transform(l, r)
            ret = be.expand(ret, expand_shape)
        elif sym == 'c': 
            ret = be.contiguous(ret)
        elif sym == 'j':
            dims = _join_transform(ret, l, r)
            ret = join(ret, dims, backend=be)
        else:
            assert False, f'Invalid transform symbol {sym}'
        if debug:
            print (f'after transform shape is: {be.shape(ret)}')
    return ret














name = "tsalib"

#import sys
#if sys.version_info < (3, 6, 1):
#    raise RuntimeError("TSAlib requires Python 3.6.1 or later")

from tsalib.ts import dim_var, dim_vars, get_dim_vars, size_assert
from tsalib.ext import view_transform, permute_transform, expand_transform, reduce_dims
from tsalib.ext import warp, join, join_transform



from sympy import symbols, Integer
from sympy import Symbol, nan, simplify
import re

def arith_op (op, s1, s2):
    assert isinstance(s1, TS)
    s2 = TS(s2)

    s1e = s1.exp
    s2e = s2.exp

    #print (f'arith_op: {op} {s1} {s2}')
    if op == 'add':
        se = s1e + s2e
    elif op == 'mul':
        se = s1e * s2e
    elif op == 'truediv':
        se = s1e / s2e  
    elif op == 'floordiv':
        se = s1e // s2e  
    else:
        raise NotImplementedError(f'{op}')

    return TS(se)

class TupleSeq:
    def __init__(self, s):
        self.s = s
    def item(self): return self.s

class DimVar:
    decls = {} #caches all dim var declarations
    parse_regexp = r'(\w+)(?:\((\w+)\))?(?::(\d+))?' #Height(h)?(:300)?

    def __init__ (self, decl, strict, cache):
        '''
        :decl: declaration string of variable ('Batch(b):20')
        :strict: allow declaration if undeclared earlier
        :cache: store in `decls` cache
        '''
        assert isinstance(decl, str)
        decl = decl.strip()

        m = re.search(DimVar.parse_regexp, decl)
        name, sname, val = m.groups()
        #print (m.groups())

        self._name = name
        self._sname = sname if sname is not None else name
        self._val = int(val) if val is not None else nan
        
        self._e = Symbol(self._sname)
        if strict and self._e in DimVar.decls:
            raise ValueError(f'DimVar {self._sname} already declared. Use strict=False to skip check.')

        if cache:       
            DimVar.decls[self._e] = self

    @property
    def exp(self): return self._e

    @property
    def len(self): return self._val

    @property
    def name(self):
        ret = f'{self._name}'
        if self._name != self._sname: ret += f'({self._sname})'
        return ret

    @staticmethod
    def check_decl(sname):
        return Symbol(sname) in DimVar.decls
    @staticmethod
    def lookup(sname):
        sn = Symbol(sname)
        assert sn in DimVar.decls, f'DimVar {sn} not declared'
        return DimVar.decls[sn]

    @staticmethod
    def eval(e):
        sub_map = [(e, dv._val) for e, dv in DimVar.decls.items()]
        ret = e.subs(sub_map)
        #print (f'eval: {e} -> {ret}')
        return ret

    @staticmethod
    def eval_name(e):
        sub_map = [(e, dv.name) for e, dv in DimVar.decls.items()]
        return str(e.subs(sub_map))

class TS:
    '''
    The Tensor Shape Expression Class
    '''
    #DEFAULT_VALUE = 1

    def __init__(self, t, is_dvar=False):
        self._e = None
        self.is_dvar = is_dvar # a basic dimension var

        if isinstance(t, int):
            self._e = Integer(t)
            self._val = t
        elif isinstance(t, DimVar):
            self._e, self._val, self.is_dvar = t.exp, t.len, True
        elif isinstance(t, TS):
            self._e, self._val, self.is_dvar = t._e, t._val, t.is_dvar
        else:
            #print (f'test expr: {v} {repr(type(v))}')
            self._e = t
            self._val = DimVar.eval(t)
            #self._val = int(v) if v is not nan else v

    @property
    def exp(self): return self._e
    @property
    def len(self): 
        return self._val if (self._val != nan) else None

    def __int__(self): 
        #print(f'called int {self._val}')
        if self._val != nan:
            return int(self._val)
        else: 
            #return TS.DEFAULT_VALUE
            raise ValueError(f'Cannot cast to integer: Default value of {self._e} not provided')
    def __index__(self): return self.__int__()

    def __add__(self, n): return arith_op('add', self, n)
    def __radd__(self, n): return self.__add__(n)
    def __mul__(self, n): return arith_op('mul', self, n)
    def __rmul__(self, n): return self.__mul__(n)

    def __floordiv__(self, n): return arith_op('floordiv', self, n)
    def __rfloordiv__(self, n): return self.__floordiv__(n)

    #truediv: '/' provided for convenience; prefer using '//'
    def __truediv__(self, n): return arith_op('truediv', self, n)
    def __rtruediv__(self, n): return self.__truediv__(n)

    def __eq__(self, d):
        #print (f'eq: {self._val}, {d}')
        if isinstance(d, int):
            #semantics: any integer matches nan
            if self._val == nan: return True 
            else: return self._val == d
        elif isinstance(d, TS):
            return self._e == d._e 
        else:
            return False   

    def __hash__(self):
        return hash(self._e)

    def __repr__(self):
        s = DimVar.eval_name(self._e)
        if self._val != nan:
            s += f':{self._val}'
        return s


def dim_var (name, strict=True, cache=True):
    '''
    Declare a single dimension variable
    '''
    d = DimVar(name, strict=strict, cache=cache)
    return TS(d)

def dummy_dvar(pos):
    '''
    Declare a dummy dimension variable at a particular dim position. Do not cache.
    '''
    assert pos >= 0
    name = f'_dm_{pos}'
    d = dim_var(name, strict=False, cache=False)
    #print (f'dummy {d}')
    return d

def dim_vars_shape(names, shape, strict=True):
    '''
    Declare dim vars corresponding to dimensions of tensor
    :names 'b t d'
    :shape (10, 30, 300)
    '''
    names = names.strip().split(' ')
    assert len(names) == len(shape), 'Number of Dimension Variables and Shape mismatch'

    tss = [dim_var(f'{name}:{shape[i]}', strict=strict) for i, name in enumerate(names)]
    if len(names) == 1: return tss[0]
    else: return tss


def dim_vars(names, strict=True, cache=True):
    '''
    Declare multiple dimension variables in one go
    '''
    names = names.strip().split(' ')
    tss = [dim_var(name, strict=strict, cache=cache) for name in names]

    if len(names) == 1: return tss[0]
    else: return tss

def get_dim_vars(names):
    '''
    names: 'b c h w', separated by spaces
    '''
    names = names.strip().split(' ')
    res = [TS(DimVar.lookup(name)) for name in names]
    if len(names) == 1: return res[0]
    else: return res

def size_assert(x_size, sa, dims=None):
    '''
    x_size: integer tuple
    sa: TSA
    dims: None or Sequence[int], e.g., [0,1]
    Check if size of tensor x matches TSA `sa` along `dims` axes
    '''
    if dims is not None:
        assert isinstance(dims, (list, tuple))
        x_size = [x_size[d] for d in dims]
        sa = [sa[d] for d in dims]

    if x_size != sa:
        print(f'Size mismatch: size = {x_size}, expected: {sa}')
        assert False


#def update_dim_var_size ():
#avoid this function
#maybe redeclare dim var?

def declare_common_dim_vars ():
    B, V, D, Dh = dim_vars('Batch Vocab EmbedDim HiddenDim')
    C, Ci, Co = dim_vars('Channels InChannels OutChannels')
    T, Te, Td = dim_vars('Time EncoderTime DecoderTime')

    return B, D, V, Dh, T, Te, Td, C, Ci, Co



#!/usr/bin/env python

import os
import setuptools

def get_long_description():
    filename = os.path.join(os.path.dirname(__file__), 'README.md')
    with open(filename) as f:
        return f.read()

setuptools.setup(name='tsalib',
      version='0.1.6.1',
      description="TSAlib: Support for Tensor Shape Annotations",
      long_description=get_long_description(),
      long_description_content_type="text/markdown",
      author='Nishant Sinha',
      author_email='nishantsinha@acm.org',
      url='https://github.com/ofnote/tsalib',
      license='Apache 2.0',
      platforms=['POSIX'],
      packages=setuptools.find_packages(),
      #entry_points={},
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Software Development',
          'Topic :: Scientific/Engineering :: Artificial Intelligence'
          ],
      setup_requires=['sympy'],
      install_requires=['sympy'],
      )